;;;-*- Mode: Lisp; Package: STELLA -*-
;;;
;;; mix worksheet
;;;
;;; **********************************************************************
;;; Copyright (c) 1995 Heinrich Taube and Tobias Kunze.  All rights
;;; reserved. Use and copying of this software and preparation of
;;; derivative works based upon this software are permitted and may be
;;; copied as long as no fees or compensation are charged for use, 
;;; copying, or accessing this software and all copies of this software
;;; include this copyright notice.
;;; **********************************************************************


(in-package :stella)


(defparameter %debug% nil)

(defun debug-format (&rest args)
  (when %debug%
    (apply #'format args)))

;;;___________________________________________________________________________
;;;
;;; mix view
;;;

(defclass mix-pane (ccl:view gray-3d-framed-view-mixin)
  ((active :initform nil :initarg :mix-pane-active :accessor mix-pane-active))
  (:default-initargs :view-nick-name ':mix-pane))

(defclass start-table (table)
  ()
  (:default-initargs
    :view-nick-name ':start-table
    :table-color *bgc*
    :view-font *capella-font*))


(defun global-p (mix-pane)
  (toggle-selected? (ccl:view-named ':gradio mix-pane)))

;;;
;;; autopositioning should go here...

(defmethod initialize-instance :after ((view mix-pane) &rest args)
  (declare (ignore args))
  ;; make a table whose first entry is the Global start time buffer
  ;; the label for this entry is invisible...
  (let ((table (make-instance 'start-table
                 :view-nick-name ':start-table
                 :view-position (xy 11 45)
                 :view-size (xy 120 0)   ; (xy (- (view-width view) 8) 0)
                 :edit-x (xy 65 0)      ; same as radios (- 73 2 3)
                 :entries (list (make-table-entry 
                                 :id ':global 
                                 :label (make-instance 'ccl:simple-view
                                          :view-size (xy 0 0)
                                          :view-position (xy 8 0)))))))
    (ccl:add-subviews 
     view
     ;(make-instance 'invisible-default-button
     ;  :action #'worksheet-action)
     (make-instance 'label
       :view-position (xy 10 16)
       :title "Start Time:")
     (make-instance 'radio-button
       :view-position (xy 73 8)
       :title "Global"
       :selected? t
       :action  #'(lambda (b) 
                    (global/local (ccl:view-container b)))
       :view-nick-name ':gradio)
     (make-instance 'radio-button
       :view-position (xy 73 24)
       :title "Local"
       :selected? nil
       :action  #'(lambda (b) 
                    (global/local (ccl:view-container b) t))
       :view-nick-name ':lradio)
     table)))

(defun get-current-reference (mix-pane)
  (let ((active '()))
    (if (and (listing-browser? *browser*) (any-selection? *browser*))
      (let ((listing (browser-listing *browser*)))
        (map-selected-objects
         #'(lambda (o) (if (event? o) (setf active nil) (push o active)))
         listing)
        (setf active (nreverse active)))
      (when *browser*
        (let ((obj (object-view-object *browser*)))
          (unless (or (systemp obj) (event? obj))
            (setf active (list obj))))))
    (setf (slot-value mix-pane 'active) active)))

(defun global/local (mix-pane &optional local?)
  (let* ((active (get-current-reference mix-pane))
         (update? nil)
         (table (ccl:view-named ':start-table mix-pane))
         (global (table-find ':global table :key #'table-entry-id)))
    (if local?
      (unless (table-entry-hidden? global)
        (hide-table-entry global)
        (let (entry)
          (dolist (e active)
            (when (setf entry (table-find e table ':key #'table-entry-id))
              (unhide-table-entry entry))))
        (setf update? t))
      (unless (table-entry-visible? global)
        (unhide-table-entry global)
        (dolist (entry (cdr (table-entries table)))
            (hide-table-entry entry))
        ;(let (entry)
        ;(dolist (e active)
        ;  (when (setf entry (table-find e table ':key #'table-entry-id))
        ;    (hide-table-entry entry))))
        (setf update? t)))
    (when update? (redisplay-changed-table table))))


(defmethod initialize-worksheet ((view mix-pane))
  (let ((active '())
        (redraw? nil)
        (global? (global-p view)))
    (setf active (get-current-reference view))
    (if active
      (let ((table (ccl:view-named ':start-table view)))
        (enable-worksheet view)
        (unless global?
          (map-table #'(lambda (e)
                         (if (member (table-entry-id e) active)
                           (when (table-entry-hidden? e)
                             (unhide-table-entry e)
                             (setf redraw? t))
                           (progn 
                             (hide-table-entry e)
                             (setf redraw? t))))
                     table))
        (dolist (a active)
          (let ((e (table-find a table :key #'table-entry-id)))
            (if e
              (unless global?
                (when (table-entry-hidden? e)
                  (setf redraw? t)))
              (progn
                (add-table-entry 
                 table (make-table-entry :id a :label (object-namestring a))
                 nil)
                (unless global? (setf redraw? t))))))
        (when redraw? (redisplay-changed-table table))
        )
      (progn 
        (unless global?
          (toggle-select (ccl:view-named ':gradio view))
          (global/local view))
        (disable-worksheet view)))
    (setf (slot-value view 'active) active)))


(defmethod disable-worksheet ((view mix-pane))
  (control-disable (ccl:view-named ':gradio view))
  (control-disable (ccl:view-named ':lradio view))
  (map-table #'(lambda (e)
                 (when (table-entry-visible? e)
                   (control-disable (table-entry-editor e))))
             (ccl:view-named ':start-table view)))


(defmethod enable-worksheet ((view mix-pane))
  (control-enable (ccl:view-named ':gradio view))
  (control-enable (ccl:view-named ':lradio view))
  (map-table #'(lambda (e)
                 (when (table-entry-visible? e)
                   (control-enable (table-entry-editor e))))
             (ccl:view-named ':start-table view)))


;;;___________________________________________________________________________
;;;
;;; mix worksheet
;;;

(defclass mix-worksheet (worksheet)
  ()
  (:default-initargs 
    :window-show nil
    :window-title "Mix"
    :color-p t))


;(defmethod ccl:view-activate-event-handler :after ((view mix-worksheet))
  ;(initialize-worksheet view)
  ;(resize-mix-pane (ccl:view-named 'mix-pane view))
  ;(redisplay-changed-table (ccl:view-named ':start-table 
  ;                                        (ccl:view-named 'mix-pane view)))
;)


(defmethod initialize-instance :after ((window mix-worksheet) &rest ignore)
  (declare (ignore ignore))
  (let* ((output-pane (make-output-stream-pane))
         (mix-pane (make-mix-pane)))
    (ccl:add-subviews window output-pane mix-pane 
                      (make-instance 'invisible-default-button
                        :action #'worksheet-action))
    (resize-mix-worksheet window)))


(defmethod initialize-worksheet ((view mix-worksheet))
  (update-stream-menu (ccl:view-named 'output-pane view))
  (initialize-worksheet (ccl:view-named 'mix-pane view)))


(defun resize-mix-worksheet (mix-worksheet)
  (let ((opane (ccl:view-named 'output-pane mix-worksheet))
        (mixpane (ccl:view-named 'mix-pane mix-worksheet))
        osize msize w)
    (when (and opane mixpane)
      (setf osize (resize-output-pane opane))
      (setf msize (resize-mix-pane mixpane))
      (setf w (max (xy-x osize) (xy-x msize)))
      ;; adjust widths of mix- and output-pane
      (when (/= (xy-x osize) (xy-y msize))
        (if (< (xy-x osize) (xy-y msize))
          (resize-output-pane opane w (xy-y osize))
          (resize-mix-pane mixpane w (xy-y msize))))
      (size-item mix-worksheet
                 (xy-x osize) (xy+ (xy-y osize) (xy-y msize)))
      (position-item mixpane 0 (xy-y osize))
      (ccl:window-show mix-worksheet))))

;;; 
;;; doesnt really autosize...

(defun autosize (view &key text font)
  (let* ((dialog-item? (or (typep view 'item)
                           (typep view 'ccl:dialog-item)))
         (old-text (if dialog-item?
                     (item-title view)
                     nil))
         (old-font (ccl:view-font view)))
    (when font 
      (ccl:set-view-font view font))
    (when (and text dialog-item?)
      (set-item-title view text))
    ;; simple-view has a dummy view-default-size method
    (ccl:set-view-size view 
                       (if dialog-item?    ;(eq (type-of view) 'ccl:simple-view)
                         (ccl:view-default-size view)
                         (let ((fnt (or font old-font)))
                           (xy (ccl:string-width text fnt)
                               (font-height fnt)))))
    (when font 
      (ccl:set-view-font view old-font))
    (when (and text dialog-item?)
      (set-item-title view old-text))
    (item-size view)))


(defun resize-mix-pane (pane &optional h v)
  (let ((table (ccl:view-named ':start-table pane)))
    (if h 
      (progn
        (multiple-value-setq (h v) (canonicalize-point-args :values h v))
        (size-item pane h v)
        (size-table table (- h 11 11)))
      (progn 
        #|
        (loop with subviews = (ccl:view-subviews 
                               (ccl:view-named ':start-table pane))
              for i from 0 to (1- (length subviews))
              for entry = (elt subviews i) 
              for label = (slot-value entry 'label)
              for editor = (slot-value entry 'editor)
              with lsize and esize
              with lxpos = 8
              with expos = 73
              with ypos = 45
              do (unless lsize
                   (setf lsize (autosize label :text "Algorithm-99")
                         esize (autosize editor :text "999")))
              (size-item label lsize)
              (size-item editor esize)
              (position-item label lxpos ypos)
              (position-item editor expos ypos)
              (incf ypos 20))
        (loop for en in (table-entries (ccl:view-named ':start-table pane))
              for l = (table-entry-label en)
              for e = (table-entry-editor en)
              do (let ((ls (ccl:view-size l))
                       (lp (ccl:view-position l))
                       (es (ccl:view-size e))
                       (ep (ccl:view-position e)))
                   (debug-format t "~&label: pos (~a ~a) size (~a ~a)~35T~
                                    textbuf: pos (~a ~a) size (~a ~a)~70T~a~75T~a"
                                 (xy-x lp) (xy-y lp) (xy-x ls) (xy-y ls) 
                                 (xy-x ep) (xy-y ep) (xy-x es) (xy-y es)
                                 (table-entry-hidden? en)
                                 (position en 
                                           (ccl:view-subviews 
                                            (ccl:view-named 
                                             ':start-table pane))))))
        |#
        (size-item pane 140 (+ 45 
                               (xy-y (item-size (ccl:view-named 
                                                 ':start-table pane)))
                               5 6))))))


(defun make-mix-pane () 
  (make-instance 'mix-pane
    :view-nick-name 'mix-pane))



(defmethod worksheet-action ((view mix-worksheet))
  ; (ccl:exit-key-handler (ccl:current-key-handler view) nil)
  (flet ((insured-time (buf)
           (multiple-value-bind (val flag) (textbuf-form buf 0)
             (if (and flag (numberp val) (>= val 0))
               val
               (progn (warn-user "~A is not a start time."
                                 (textbuf-text buf))
                      (return-from worksheet-action nil))))))
    (let* ((objects '())
           (starts '())
           (mixpane (ccl:view-named 'mix-pane view))
           (table (ccl:view-named ':start-table mixpane))
           (global? (global-p mixpane)))
      (catch :abort
        (multiple-value-bind (stream syntax type text)
                             (output-spec-to-stream-object
                              (ccl:view-named 'output-pane view))
          (declare (ccl:ignore-if-unused stream)) 
          ;(format t "~%~A ~A ~A ~A"  stream syntax type text)
          (if global?
            (setf objects (get-current-reference 
                           (ccl:view-named 'mix-pane view))
                  ;(slot-value view 'active)
                  starts (insured-time 
                          (table-entry-editor
                           (table-find ':global table :key #'table-entry-id))))
            (progn
              (map-table #'(lambda (e)
                             (when (table-entry-visible? e)
                               (push (table-entry-id e) objects)
                               (push (insured-time (table-entry-editor e))
                                     starts)))
                         table)
              (setf objects (nreverse objects) starts (nreverse starts))))
          
          (let ((*syntax* syntax)
                (*command-prompting* nil))
            (when objects
              (debug-format t "~%mixing: ~S ~S"  objects starts)
              (ccl:with-cursor ccl:*watch-cursor*
                (if (eq type :file)
                  (write-object objects text starts))
                (listen-object objects starts)))))))))


#|
(make-instance 'mix-worksheet)

(ccl:window-select (make-instance 'mix-pane :view-size (xy 150 75)
                                :window-show nil))
|#

;;;
;;; -*- EOF -*-